package com.install4j.api.formcomponents;

import com.install4j.api.beans.AbstractBean;
import com.install4j.api.context.Context;
import com.install4j.api.context.UserCanceledException;
import com.install4j.api.screens.Console;
import com.install4j.runtime.beans.formcomponents.FormEnvironmentContainer;
import com.install4j.runtime.beans.formcomponents.FormEnvironmentImpl;
import com.install4j.runtime.installer.frontend.FormPanel;
import com.install4j.runtime.installer.frontend.ScreenEnvelope;

import javax.swing.*;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Abstract base class for form components.
 * All methods that have a reasonable default answer are overridden in the class.
 * This class saves the context and the form environment that are set by the framework and provides getters for them.
 * In addition the "enabled" and "visible" properties are handled by invoking the corresponding setters on all
 * components in the {@code ComponentTuple} of the form component.
 *
 * @author ej-technologies GmbH
 */
public abstract class AbstractFormComponent extends AbstractBean implements FormComponent, FormEnvironmentContainer {

    public static final Pattern MNEMONIC_PATTERN = Pattern.compile("&(\\w)");

    public static String getTextWithoutMnemonics(String text) {
        return MNEMONIC_PATTERN.matcher(text).replaceFirst("$1");
    }

    public static int getMnemonicCharIndex(String text) {
        Matcher matcher = MNEMONIC_PATTERN.matcher(text);
        if (matcher.find()) {
            return matcher.start();
        } else {
            return -1;
        }
    }
    
    private boolean enabled = true;
    private Context context;
    private FormEnvironment formEnvironment;
    private boolean visible = true;

    @Override
    public void setContext(Context context) {
        this.context = context;
    }

    /**
     * Returns the {@code Context} that the framework has associated with
     * this form component in {@code setContext}.
     * @return the installer context.
     */
    public Context getContext() {
        return context;
    }

    @Override
    public void setFormEnvironment(FormEnvironment formEnvironment) {
        this.formEnvironment = formEnvironment;
    }

    /**
     * Returns the {@code FormEnvironment} that the framework has associated with
     * this form component in {@code setFormEnvironment}.
     * @return the installer context.
     */
    @Override
    public FormEnvironment getFormEnvironment() {
        return formEnvironment;
    }

    @Override
    public boolean isEnabled() {
        return enabled;
    }

    @Override
    public void setEnabled(boolean enabled) {
        this.enabled = enabled;
        Object configurationObject = getConfigurationObject();
        if (configurationObject instanceof JComponent) {
            ((JComponent)configurationObject).setEnabled(enabled);
        }
        if (formEnvironment instanceof FormEnvironmentImpl) {
            FormPanel formPanel = ((FormEnvironmentImpl)formEnvironment).getFormPanel();
            ScreenEnvelope screenEnvelope = formPanel.getScreenEnvelope();
            if (screenEnvelope != null) {
                screenEnvelope.formComponentEnabledChanged(this, enabled);
            }
        }
    }

    @Override
    public void setVisible(boolean visible) {
        this.visible = visible;
        if (formEnvironment != null) {
            ComponentTuple componentTuple = formEnvironment.getComponentTuple(this);
            if (componentTuple != null) {
                setVisible(componentTuple.getLeftComponent(), visible);
                setVisible(componentTuple.getCenterComponent(), visible);
                setVisible(componentTuple.getRightComponent(), visible);
            }
        }
    }

    private void setVisible(JComponent component, boolean visible) {
        if (component != null) {
            component.setVisible(visible);
        }
    }

    @Override
    public boolean isVisible() {
        return visible;
    }

    @Override
    public void migrateIds(Map<String, String> oldItToNewId) {

    }

    @Override
    public void requestFocus() {
        Object configurationObject = getConfigurationObject();
        if (configurationObject instanceof JComponent) {
            final JComponent component = (JComponent)configurationObject;
            if (component.isFocusable()) {
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        component.requestFocus();
                    }
                });
            }
        }
    }

    @Override
    public boolean hasUserInput() {
        return true;
    }

    @Override
    public JComponent createLeftComponent() {
        return null;
    }

    @Override
    public boolean isFillCenterVertical() {
        return false;
    }

    @Override
    public JComponent createRightComponent() {
        return null;
    }

    @Override
    public Object getConfigurationObject() {
        return null;
    }

    @Override
    public Class getConfigurationObjectClass() {
        return null;
    }

    @Override
    public boolean checkCompleted() {
        return true;
    }

    @Override
    public void initialize() {
    }

    // catch overridden methods with this name during compilation
    /**
     * The name of this method has changed to "initialize"
     */
    protected final void initalize() {
    }

    @Override
    public void formWillActivate() {
    }

    @Override
    public void formActivated() {
    }

    @Override
    public void formDeactivated() {
    }

    @Override
    public boolean handleConsole(Console console) throws UserCanceledException {
        return true;
    }

    @Override
    public boolean handleUnattended() {
        return true;
    }

    /**
     * Helper method to get an init value. If the variable is defined and of class valueClass, the
     * variable value is used. Otherwise the default value is used.
     * @param defaultValue the default value. Can be null.
     * @param variableName the installer variable name the value will be assigned to later on
     * @param valueClass the desired class of the variable content
     * @return the init value to be used
     */
    protected Object getInitValue(Object defaultValue, String variableName, Class<?> valueClass) {
        Context context = getContext();
        Object value = context == null ? null : context.getVariable(variableName);
        if (value != null && valueClass.isAssignableFrom(value.getClass())) {
            return value;
        } else {
            return defaultValue;
        }
    }
}
